#include "data_struct.h"
#include "ethread.h"
#include <sys/socket.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#ifndef __USE_GNU
 #define __USE_GNU
 #include <fcntl.h>
 #undef __USE_GNU
#else
 #include <fcntl.h>
#endif
#include <string.h>
#include <sys/poll.h>
#include <assert.h>
#include <signal.h>
#include <errno.h>
#ifdef _USE_FDM_LIBRARY
#include "../fdm/fdm.h"
#endif

#define STD_MESSAGE_SIZE		128
#define STD_SERVER_PORT			8080
#define DATA_BUFFER_SIZE		2048
#define MIN_AHEAD_SPACE			(DATA_BUFFER_SIZE / 12)
#define SEC_TO_USEC 1000000

#if defined(_USE_RTSIGNALS)
int rtsig;
#endif

static int msgsize = STD_MESSAGE_SIZE, port = STD_SERVER_PORT;
extern int stksize;
extern queue_t *ready_queue;
queue_t *backup;
struct timeval t1, t2;
int abort_server = 0;

#ifdef _USE_RTSIGNALS
extern int sigio_cnt;
extern int sig_cnt;
#endif

int open_socket(int domain, int type, int protocol)
{
	int sfd = socket(domain, type, protocol),
	flags = 1;
#ifdef _USE_FDM_LIBRARY
	struct pollfd regfd;
#endif

	if (sfd == -1)return -1;

#if defined(_USE_RTSIGNALS)
	if ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
			fcntl(sfd, F_SETFL, flags | O_NONBLOCK | O_ASYNC) < 0) {
		close(sfd);
		return -1;
	}
	fcntl(sfd, F_SETSIG, rtsig);
	fcntl(sfd, F_SETOWN, getpid());
#else
	if (ioctl(sfd, FIONBIO, &flags) &&
			((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
			fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0)) {
		close(sfd);
		return -1;
	}
#ifdef _USE_FDM_LIBRARY
	regfd.fd = sfd;
	regfd.events = POLLIN;
	fdm_reg_fd(&regfd, 1);
#endif
#endif

	return sfd;
}


void usage(char const *prgname)
{
	fprintf(stderr,"use: %s [--msgsize nbytes (%d)] [--port nbr (%d)]\n", prgname, msgsize, port);
}


int _read(int conn, char *buf, int nbyte)
{
	int n, ret;

	while ((n = read(conn, buf, nbyte)) < 0) {
		if (errno == EINTR)
			continue;
		if ((errno != EAGAIN) && (errno != EWOULDBLOCK))
			return -1;
		ret = sleep_on_fd(conn, POLLIN, 1000);
	}
	return n;
}


int _write(int conn, char const *buf, int nbyte)
{
	int n, ret;

	while ((n = write(conn, buf, nbyte)) < 0) {
		if (errno == EINTR)
			continue;
		if ((errno != EAGAIN) && (errno != EWOULDBLOCK))
			return -1;
		ret = sleep_on_fd(conn, POLLOUT, 1000);	
	}
	assert (n == nbyte);
	return n;
}


/* return -1 means real ERROR occurs. this function exclude EAGAIN.., etc */
int _accept(int conn, struct sockaddr *addr, int *addrlen)
{
	int sfd, flags = 1, ret;
#ifdef _USE_FDM_LIBRARY
	struct pollfd regfd;
#endif

	while ((sfd = accept(conn, addr, (socklen_t *) addrlen)) < 0) {
		if (errno == EINTR)
			continue;
		if ((errno != EAGAIN) && (errno != EWOULDBLOCK))
			return -1;
		ret = sleep_on_fd(conn, POLLIN, 1000);
		if (ret < 0 && ret != POLL_TIMEOUT) {
			printf("WARNING: _accept error on sleep_on_fd (ret = %d\n).", ret);
		}
	}
#if defined(_USE_RTSIGNALS)
	if ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
			fcntl(sfd, F_SETFL, flags | O_NONBLOCK | O_ASYNC) < 0) {
		close(sfd);
		return -1;
	}
	fcntl(sfd, F_SETSIG, rtsig);
	fcntl(sfd, F_SETOWN, getpid());
#else
	if (ioctl(sfd, FIONBIO, &flags) &&
			((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
			fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0)) {
		close(sfd);
		return -1;
	}
#ifdef _USE_FDM_LIBRARY
	regfd.fd = sfd;
	regfd.events = POLLIN;
	fdm_reg_fd(&regfd, 0);
#endif
#endif
	return sfd;
}


static int read_data(int conn, int *rindex, int *nbytes, char buffer[])
{
	int nb;

	if (*rindex > (DATA_BUFFER_SIZE - MIN_AHEAD_SPACE)) {
		memcpy(buffer, buffer + *rindex, *nbytes - *rindex);
		*nbytes = *nbytes - *rindex;
		*rindex = 0;
	} else if (*nbytes == DATA_BUFFER_SIZE && *rindex > 0) {
		if (*rindex > (DATA_BUFFER_SIZE >> 1))
			memcpy(buffer, buffer + *rindex, *nbytes - *rindex);
		else
			memmove(buffer, buffer + *rindex, *nbytes - *rindex);
		*nbytes = *nbytes - *rindex;
		*rindex = 0;
	}

	if ((nb = _read(conn, buffer + *rindex,
			DATA_BUFFER_SIZE - *rindex)) <= 0)
		return -1;

	*nbytes += nb;

	return 0;
}


static char *read_line(int conn, int *rindex, int *nbytes, char buffer[])
{
	char *nline = NULL;

	for (; !nline;) {
		if (*nbytes > *rindex) {
			if ((nline = memchr(buffer + *rindex, '\n', *nbytes - *rindex))) {
				char *line = buffer + *rindex;

				*rindex += (nline - line) + 1;
				for (; nline > line && nline[-1] == '\r'; nline--);
				*nline = '\0';
				return line;
			}
			if (*nbytes == DATA_BUFFER_SIZE)
				break;
		}
		if (read_data(conn, rindex, nbytes, buffer) < 0)
			break;
	}
	return NULL;
}


static int parse_request(int conn, int *rindex, int *nbytes, char buffer[])
{
	char *line;

	if (!(line = read_line(conn, rindex, nbytes, buffer)))
		return -1;

//	printf("line: %s\n", line);

	for (;;) {
		if (!(line = read_line(conn, rindex, nbytes, buffer)))
			return -1;
//		printf("line: %s\n", line);
		if (*line == '\0')
			break;


	}
	return 0;
}


static int send_response(int conn)
{
	static int resplen = -1;
	static char *resp = NULL;
	int ret;

	if (resp == NULL) {
		msgsize = ((msgsize + 63) / 64) * 64;

		resp = (char *) malloc(msgsize + 256);

		sprintf(resp,
				"HTTP/1.1 200 OK\r\n"
				"Server: dp server\r\n"
				"Content-Type: text/plain\r\n"
				"Content-Length: %d\r\n"
				"\r\n", msgsize);

		while (msgsize > 0) {
			strcat(resp, "01234567890123\r\n" "01234567890123\r\n" "01234567890123\r\n" "01234567890123\r\n");
			msgsize -= 64;
		}

		resplen = strlen(resp);
	}

	ret = _write(conn, resp, resplen);
	if (ret == -1) return -1;

	return 0;
}


static void *httpd(void *data)
{
	int conn = (int) data;
	int nbytes = 0;
	int rindex = 0;
	char buffer[DATA_BUFFER_SIZE];
	int ret;

	while (parse_request(conn, &nbytes, &rindex, buffer) == 0) {
		ret = send_response(conn);
		if (ret != 0) {	
			perror("send_response error!\n");
		}
	}
#ifdef _USE_FDM_LIBRARY
	ret = fdm_unreg_fd(conn);
	assert(ret == 0);
#endif
//	printf("fd %d closed\n", conn);
	close(conn);
	return data;
}


static void *acceptor(void *data)
{
	int conn_socket = (int)data;
	struct sockaddr_in addr;
	int ret, sfd, addrlen = sizeof(addr);

	while ((sfd = _accept(conn_socket, (struct sockaddr *) &addr, &addrlen)) != -1) {
		ret = ethread_create(httpd, (void *)sfd);
#ifdef _NO_MULTI_ACCEPT
		thread_yield();
#endif		
		if (ret < 0) {
#ifdef _USE_FDM_LIBRARY
			fdm_unreg_fd(sfd);
#endif	
			close(sfd);
			return (void *)ret;
		}
	}
	return data;
}


void timerec(int val) {
	abort_server = 1;  
}


int main(int argc, char *argv[]) {

	int ii, ret, sfd, flags = 1;
	struct linger ling = {0, 0};
	struct sockaddr_in addr;
#if defined(_USE_RTSIGNALS)
	sigset_t sset;
#endif
	struct rusage r;
#ifdef _USE_THREADS
	struct rusage r0, r1, r2;
	long u0time, s0time, u1time, s1time, u2time, s2time;
#endif
	float p;
	long etime, utime, stime;


	for (ii = 1; ii < argc; ii++) {
		if (strcmp(argv[ii], "--msgsize") == 0) {
			if (++ii < argc)
				msgsize = atoi(argv[ii]);
			continue;
		}
		if (strcmp(argv[ii], "--port") == 0) {
			if (++ii < argc)
				port = atoi(argv[ii]);
			continue;
		}

		usage(argv[0]);
		return -1;
	}

	gettimeofday(&t1, NULL);	
	signal(SIGINT, timerec);

#if defined(_USE_RTSIGNALS)
	rtsig = SIGRTMIN + 1;

	sigemptyset(&sset);
	sigaddset(&sset, rtsig);
	sigaddset(&sset, SIGIO);
	sigprocmask(SIG_BLOCK, &sset, NULL);
#endif

	ethread_init();

	backup = ready_queue;
   
	if ((sfd = open_socket(AF_INET, SOCK_STREAM, 0)) == -1) {	
		perror("listen socket open fail!");
		return -1;
	}

	setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));
	setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, &flags, sizeof(flags));
	setsockopt(sfd, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling));

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(sfd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
		close(sfd);
		return -1;
	}

	listen(sfd, 128);

	ret = ethread_create(acceptor, (void *)sfd);
	if (ret < 0) return -1;

	while (!abort_server) {
		getCurrentTime();
		scheduler();
	}

	ethread_shutdown();


	/* report user time, system time and CPU usage */

	gettimeofday(&t2, NULL);
	getrusage(RUSAGE_SELF, &r);
#ifdef _USE_THREADS
	fdm_getrusage(0, &r0);
	fdm_getrusage(1, &r1);
	fdm_getrusage(2, &r2);
#endif
	etime = (t2.tv_sec - t1.tv_sec)*SEC_TO_USEC+(t2.tv_usec-t1.tv_usec);
	utime = r.ru_utime.tv_sec*SEC_TO_USEC+r.ru_utime.tv_usec;
	stime = r.ru_stime.tv_sec*SEC_TO_USEC+r.ru_stime.tv_usec;
        printf("user time = %ld\n", utime);
        printf("system time = %ld\n", stime);
#ifdef _USE_RTSIGNALS
	printf("sigio count = %d\n", sigio_cnt);
	printf("sig count = %d\n", sig_cnt);
#endif        


#ifdef _USE_THREADS
	u0time = r0.ru_utime.tv_sec*SEC_TO_USEC+r0.ru_utime.tv_usec;
	s0time = r0.ru_stime.tv_sec*SEC_TO_USEC+r0.ru_stime.tv_usec;
	u1time = r1.ru_utime.tv_sec*SEC_TO_USEC+r1.ru_utime.tv_usec;
	s1time = r1.ru_stime.tv_sec*SEC_TO_USEC+r1.ru_stime.tv_usec;
	u2time = r2.ru_utime.tv_sec*SEC_TO_USEC+r2.ru_utime.tv_usec;
	s2time = r2.ru_stime.tv_sec*SEC_TO_USEC+r2.ru_stime.tv_usec;
        printf("thread 0 user time = %ld\n", u0time);
        printf("thread 0 system time = %ld\n", s0time);
        printf("thread 1 user time = %ld\n", u1time);
        printf("thread 1 system time = %ld\n", s1time);        
        printf("thread 2 user time = %ld\n", u2time);
        printf("thread 2 system time = %ld\n", s2time);        
#endif
        printf("elapsed time = %ld\n", etime);
#ifdef _USE_THREADS
	p = (float) (utime+stime+u0time+s0time+u1time+s1time+u2time+s2time) / etime;
#else	
	p = (float) (utime + stime) / etime;
#endif
	printf("cpu usage = %f\n", p*100);

	return 0;
}
